import argparse
import sys
from pathlib import Path, PurePosixPath

template = """
python_sources = [
{sources}
]

py.install_sources(
  python_sources,
  subdir: '{subdir}'
)
"""


def process_directory(path_dir, path_pack=None):

    if path_pack is None:
        path_pack = path_dir
        print(f"Process {path_dir}")
        subdir = path_pack.name
    else:
        subdir = PurePosixPath(path_dir.relative_to(path_pack.parent))
        print(f"Process subdir {subdir}")

    names_py = sorted(path.name for path in path_dir.glob("*.py"))
    print(f"{names_py = }")
    paths_subdirs = [
        path
        for path in path_dir.glob("*")
        if path.is_dir() and not path.name.startswith("__")
    ]

    names_subdirs = sorted(path.name for path in paths_subdirs)
    print(f"{names_subdirs = }")

    if names_py:
        code = template.format(
            sources="\n".join(f"  '{name}'," for name in names_py),
            subdir=subdir,
        )
    else:
        code = ""

    if names_subdirs:
        code += (
            "\n" + "\n".join(f"subdir('{name}')" for name in names_subdirs) + "\n"
        )

    if not code:
        return

    with open(path_dir / "meson.build", "w", encoding="utf-8") as file:
        file.write(code)

    for path_subdir in paths_subdirs:
        process_directory(path_subdir, path_pack)


def main():

    parser = argparse.ArgumentParser(
        prog="transonic-init-meson",
        description="Create Meson files from a Python package",
    )
    parser.add_argument("path", help="Path package.")

    args = parser.parse_args()

    path_pack = Path(args.path)

    if not path_pack.is_dir():
        print("Path given is not a directory", file=sys.stderr)
        sys.exit(1)

    process_directory(path_pack)



"""Tests the Meson support

"""

import os
import subprocess
import sys
from pathlib import Path
from shutil import copy
from unittest.mock import patch

import pytest

from transonic.config import backend_default
from transonic.mpi import nb_proc
from transonic.run import run

test_dir = Path(__file__).absolute().parent

path_root_package = test_dir.parent / "data_tests/package_for_test_meson"
assert path_root_package.exists()

path_src_package = path_root_package / "src/package_for_test_meson"
assert path_src_package.exists()


def run_in_venv(venv, command, cwd=None):
    try:
        env = venv.env
    except AttributeError:
        if sys.platform.startswith("win"):
            name_dir_executables = "Scripts"
            sep_path_var = ";"
        else:
            name_dir_executables = "bin"
            sep_path_var = ":"
        path_bin = Path(venv.path) / name_dir_executables
        env = os.environ.copy()
        env["PATH"] = str(path_bin) + sep_path_var + os.environ["PATH"]
        venv.env = env

    args = [venv.python, "-m"]
    args.extend(command.split())
    subprocess.run(args, cwd=cwd, check=True, env=env)


def eval_in_venv(venv, code):
    process = subprocess.run(
        [venv.python, "-c", code],
        env=venv.env,
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )
    return process.stdout


@pytest.mark.xfail(
    sys.platform.startswith("win"), reason="Buggy on Windows (TODO: debug)"
)
@pytest.mark.skipif(nb_proc > 1, reason="No commandline in MPI")
@pytest.mark.xfail(
    backend_default not in ["pythran", "python"], reason="Not yet implemented"
)
def test_install_package(tmpdir, venv):
    """
    Test the installation of the package data_tests/package_for_test_meson

    - create and activate a temporary virtual environment
    - create a clean package_for_test_meson in tmp directory
    - install transonic from source
    - install other build requirements (meson-python and ninja) and pytest
    - install pythran for the pythran backend
    - install package_for_test_meson with `pip install . --no-build-isolation` (+ give the backend)
    - test package_for_test_meson with `pytest tests`

    """

    assert venv.python.endswith("python")

    for name in ("pyproject.toml", "meson.build", "meson.options", "README.md"):
        copy(path_root_package / name, tmpdir)

    src_dir = path_root_package / "tests"
    dst_dir = tmpdir / "tests"
    dst_dir.mkdir()
    for name in ("test_bar.py", "test_foo.py"):
        copy(src_dir / name, dst_dir)

    src_dir = path_src_package
    dst_dir = Path(tmpdir / "src/package_for_test_meson")
    dst_dir.mkdir(parents=True)
    for name in ("bar.py", "foo.py", "meson.build", "__init__.py"):
        copy(src_dir / name, dst_dir)

    run_in_venv(venv, f"pip install {test_dir.parent}")
    run_in_venv(venv, "pip install meson-python meson ninja pytest")

    if backend_default == "pythran":
        run_in_venv(venv, "pip install pythran")

    install_command = "pip install -e . --no-build-isolation"
    if backend_default == "python":
        install_command += (
            " --config-settings=setup-args=-Dtransonic-backend=python"
        )

    run_in_venv(venv, install_command, cwd=tmpdir)
    run_in_venv(venv, "pytest tests", cwd=tmpdir)

    out = eval_in_venv(
        venv, "from package_for_test_meson.foo import ts; print(ts.is_compiled)"
    ).strip()
    is_compiled = out == "True"
    assert is_compiled


@pytest.mark.skipif(nb_proc > 1, reason="No commandline in MPI")
@pytest.mark.xfail(backend_default == "cython", reason="Not yet implemented")
def test_meson_option(tmpdir, monkeypatch):
    """Only run `transonic --meson foo.py bar.py` in
    data_tests/package_for_test_meson/src/package_for_test_meson
    and compare with for_test__pythran__meson.build (same directory)
    """

    for name in ("bar.py", "foo.py", "meson.build"):
        copy(path_src_package / name, tmpdir)

    monkeypatch.chdir(tmpdir)

    argv = "transonic --meson bar.py foo.py".split()
    with patch.object(sys, "argv", argv):
        run()

    path_result = tmpdir / f"__{backend_default}__/meson.build"
    assert path_result.exists()

    saved_output_path = (
        path_src_package / f"for_test__{backend_default}__meson.build"
    )
    assert saved_output_path.exists()

    saved_meson_build = saved_output_path.read_text()
    produced_meson_build = path_result.read_text("utf8")

    assert saved_meson_build == produced_meson_build



"""Jax backend
================

Internal API
------------

.. autoclass:: SubBackendJITJax
   :members:
   :private-members:

.. autoclass:: JaxBackend
   :members:
   :private-members:

"""

from typing import Optional

from transonic.analyses.extast import CommentLine, gast, parse, unparse
from transonic.util import format_str

from .py import PythonBackend, SubBackendJITPython


def add_jax_comments(code):
    """Add Jax code in Python comments"""
    mod = parse(code)
    new_body = [CommentLine("# __protected__ from jax import jit")]

    for node in mod.body:
        # Replace `import numpy` -> `import jax.numpy as numpy`
        # Replace `import numpy as np` -> `import jax.numpy as np`
        if isinstance(node, gast.Import):
            if (alias := node.names[0]).name == "numpy":
                g_alias = gast.alias(
                    name="jax.numpy", asname=alias.asname or alias.name
                )
                node = gast.Import([g_alias])

        # Replace `from numpy import eye` -> `from jax.numpy import eye`
        elif isinstance(node, gast.ImportFrom):
            if node.module == "numpy":
                node.module = "jax.numpy"

        # Add JIT decorator
        if isinstance(node, gast.FunctionDef):
            new_body.append(CommentLine("# __protected__ @jit"))
        new_body.append(node)

    mod.body = new_body
    return format_str(unparse(mod))


class SubBackendJITJax(SubBackendJITPython):
    def make_backend_source(self, info_analysis, func, path_backend):
        src, has_to_write = super().make_backend_source(
            info_analysis, func, path_backend
        )

        if not src:
            return src, has_to_write

        return add_jax_comments(src), has_to_write


class JaxBackend(PythonBackend):
    """Main class for the Jax backend

    See https://github.com/google/jax

    """

    backend_name = "jax"
    _SubBackendJIT = SubBackendJITJax

    def compile_extension(
        self,
        path_backend,
        name_ext_file=None,
        native=False,
        xsimd=False,
        openmp=False,
        str_accelerator_flags: Optional[str] = None,
        parallel=True,
        force=True,
    ):
        if name_ext_file is None:
            name_ext_file = self.name_ext_from_path_backend(path_backend)

        with open(path_backend) as file:
            source = file.read()

        source = source.replace("# __protected__ ", "")

        with open(path_backend.with_name(name_ext_file), "w") as file:
            file.write(format_str(source))

        compiling = False
        process = None
        return compiling, process

    def _make_backend_code(self, path_py, analysis, **kwargs):
        """Create a backend code from a Python file"""
        code, codes_ext, header = super()._make_backend_code(path_py, analysis)

        if not code:
            return code, codes_ext, header

        code = add_jax_comments(code)

        for_meson = kwargs.get("for_meson", False)
        if for_meson:
            code = format_str(code.replace("# __protected__ ", ""))

        return code, codes_ext, header



import sys
from pathlib import PurePosixPath
from unittest.mock import patch

from transonic_cl.init_meson import main

cmd = "transonic-init-meson"


mypack = {
    "mypack": {
        "a.py": None,
        "b.py": None,
        "c.txt": None,
        "subpack0": {"d.py": None},
        "subpack1": {},
        "subpack2": {"subsubpack": {"e.py": None}},
    }
}


meson_files = {
    "mypack": """
python_sources = [
  'a.py',
  'b.py',
]

py.install_sources(
  python_sources,
  subdir: 'mypack'
)

subdir('subpack0')
subdir('subpack1')
subdir('subpack2')
""",
    "mypack/subpack0": """
python_sources = [
  'd.py',
]

py.install_sources(
  python_sources,
  subdir: 'mypack/subpack0'
)
""",
    "mypack/subpack2": """
subdir('subsubpack')
""",
    "mypack/subpack2/subsubpack": """
python_sources = [
  'e.py',
]

py.install_sources(
  python_sources,
  subdir: 'mypack/subpack2/subsubpack'
)
""",
}


def create_package(package: dict, path_dir):
    path_dir.mkdir(exist_ok=True)
    for key, value in package.items():
        if value is None:
            # it should be a file
            (path_dir / key).touch()
        else:
            assert isinstance(value, dict)
            create_package(value, path_dir / key)


def test_init_meson(tmp_path):
    path_mypack = tmp_path / "mypack"
    print(f"create package\n{path_mypack}")
    create_package(mypack, tmp_path)

    with patch("sys.argv", [cmd, str(path_mypack)]):
        main()

    meson_paths = sorted(path_mypack.rglob("meson.build"))
    assert len(meson_paths) == len(meson_files)

    for meson_path in meson_paths:
        code = (path_mypack / meson_path).read_text(encoding="utf-8")

        meson_path = PurePosixPath(meson_path.relative_to(path_mypack.parent))
        print(meson_path)
        code_should_be = meson_files[str(meson_path.parent)]

        if code != code_should_be:
            print(code_should_be)

        assert code == code_should_be



import argparse
import sys
from importlib import import_module
from pathlib import Path


def main():
    parser = argparse.ArgumentParser(
        prog="transonic-get-include",
        description="Get include directory for packages",
    )

    parser.add_argument("package", type=str, help="Package name")

    args = parser.parse_args()

    try:
        mod = import_module(args.package)
    except ImportError:
        print("ImportError")
        sys.exit(1)

    try:
        path_include = Path(mod.get_include())
    except AttributeError:
        print(f"No {args.package}.get_include")
        sys.exit(1)

    try:
        path_include = path_include.relative_to(Path.cwd())
    except ValueError:
        pass

    print(path_include)



import os
import sys
from shutil import rmtree
from time import sleep

import numpy as np
import pytest

from transonic import mpi
from transonic.backends import backends
from transonic.compiler import scheduler, wait_for_all_extensions
from transonic.config import backend_default
from transonic.justintime import modules
from transonic.util import can_import_accelerator

backend = backends[backend_default]
scheduler.nb_cpus = 2

module_name = "_transonic_testing.for_test_justintime"

str_relative_path = module_name.replace(".", os.path.sep)

path_jit = backend.jit.path_base
path_jit_classes = backend.jit.path_class

path_jit_dir = path_jit / str_relative_path
path_classes_dir = path_jit_classes / str_relative_path
path_classes_dir1 = path_jit / path_jit_classes.name / str_relative_path

if mpi.rank == 0:
    if path_jit.exists():
        rmtree(path_jit_dir, ignore_errors=True)
    if path_classes_dir.exists():
        rmtree(path_classes_dir)
    if path_classes_dir1.exists():
        rmtree(path_classes_dir1)

mpi.barrier()


@pytest.mark.skipif(backend_default == "numba", reason="Not supported by Numba")
def test_jit():
    from _transonic_testing.for_test_justintime import func1

    a = np.arange(2)
    b = [1, 2]

    for _ in range(2):
        func1(a, b)
        sleep(0.1)

    wait_for_all_extensions()
    func1(a, b)


def test_fib():
    from _transonic_testing.for_test_justintime import fib, use_fib

    fib2 = fib(2)
    result = use_fib()
    wait_for_all_extensions()
    assert fib2 == fib(2)
    assert result == use_fib()

    fib15 = fib(1.5)
    wait_for_all_extensions()
    assert fib15 == fib(1.5)


def test_jit_simple():
    from _transonic_testing.for_test_justintime import func_identity

    func_identity(1)

    if not can_import_accelerator():
        return

    mod = modules[module_name]
    cjit = mod.jit_functions["func_identity"]

    for _ in range(100):
        func_identity(1)
        sleep(0.1)
        if not cjit.compiling:
            sleep(0.1)
            func_identity(1)
            break

    del sys.modules[module_name]
    del modules[module_name]

    from _transonic_testing.for_test_justintime import func_identity

    func_identity(1)


def test_jit_array_bool():
    from _transonic_testing.for_test_justintime import func_identity

    a = np.zeros(2, dtype=bool)
    func_identity(a)
    wait_for_all_extensions()
    assert np.allclose(a, func_identity(a))


@pytest.mark.skipif(backend_default == "numba", reason="Not supported by Numba")
def test_jit_dict():
    from _transonic_testing.for_test_justintime import func_dict

    d = dict(a=1, b=2)
    func_dict(d)

    if not can_import_accelerator():
        return

    mod = modules[module_name]
    cjit = mod.jit_functions["func_dict"]

    d = dict(a=1, b=2)
    func_dict(d)

    wait_for_all_extensions()

    for _ in range(100):
        d = dict(a=1, b=2)
        func_dict(d)
        sleep(0.1)
        if not cjit.compiling:
            sleep(0.1)
            func_dict(d)
            break


def test_jit_method():
    from _transonic_testing.for_test_justintime import MyClass

    obj = MyClass()
    obj.check()

    if not can_import_accelerator():
        return

    obj = MyClass()
    obj.check()

    wait_for_all_extensions()

    obj.check()


@pytest.mark.skipif(backend_default == "numba", reason="Not supported by Numba")
def test_jit_method2():
    from _transonic_testing.for_test_justintime import MyClass2

    obj = MyClass2()
    obj.check()

    if not can_import_accelerator():
        return

    obj = MyClass2()
    obj.check()

    wait_for_all_extensions()

    obj.check()


def test_func0():
    from _transonic_testing.for_test_justintime import func0, func0_jitted

    func02 = func0(2.1)
    result = func0_jitted(2.1)
    wait_for_all_extensions()
    assert func02 == func0(2.1)
    assert result == func0(2.1)


# jitted function that uses a local function, a jitted local function
# and a jitted imported function (with a different name)
def test_main():
    from _transonic_testing.for_test_justintime import main

    main_res = main(2)
    wait_for_all_extensions()
    assert main_res == 5


def test_jit_imported():
    from _transonic_testing.for_test_justintime import (
        func_import,
        jitted_func_import,
    )

    result = jitted_func_import()
    wait_for_all_extensions()
    assert result == jitted_func_import() == func_import()



from unittest.mock import patch

from transonic.backends import backends
from transonic_cl.clean_dir import main

cmd = "transonic-clean-dir"


names = [backend for backend in backends]


def test_clean_dir(tmp_path):

    subpack_foo = tmp_path / "foo"
    subpack_bar = subpack_foo / "bar"

    subpack_bar.mkdir(parents=True)

    for subpack in (subpack_foo, subpack_bar):
        for name in names:
            (subpack / f"__{name}__").mkdir()

    with patch("sys.argv", [cmd, "-h"]):
        try:
            main()
        except SystemExit as system_exit:
            assert system_exit.code == 0
        else:
            raise RuntimeError("Issue with -h")

    with patch("sys.argv", [cmd, str(tmp_path / "does_not_exist")]):
        try:
            main()
        except SystemExit as system_exit:
            assert system_exit.code == 1
        else:
            raise RuntimeError("Issue with not is_dir()")

    with patch("sys.argv", [cmd, str(tmp_path)]):
        main()

    for subpack in (subpack_foo, subpack_bar):
        assert subpack.exists()
        for name in names:
            assert not (subpack / f"__{name}__").exists()



import pytest

from transonic_cl.get_includes import main


class MyException(Exception):
    pass


def test_get_include(monkeypatch):
    monkeypatch.setattr("sys.argv", ["transonic-get-include", "numpy"])
    main()

    def _exit(exit_code):
        if exit_code == 1:
            raise MyException

    monkeypatch.setattr("sys.exit", _exit)
    monkeypatch.setattr("sys.argv", ["transonic-get-include", "numpyyyy"])

    with pytest.raises(MyException):
        main()



import argparse
import sys
from pathlib import Path
from shutil import rmtree

from transonic.backends import backends

cmd = "transonic-clean-dir"


names = set(f"__{backend}__" for backend in backends)


def clean_dir(path_dir):
    """Delete backend files from a directory (recursive)"""
    subdirs_all = sorted(path for path in path_dir.glob("*") if path.is_dir())
    subdirs = []
    for subdir in subdirs_all:
        if subdir.name in names:
            rmtree(subdir, ignore_errors=True)
            continue
        subdirs.append(subdir)
    for subdir in subdirs:
        clean_dir(subdir)


def main():

    parser = argparse.ArgumentParser(
        prog="transonic-clean-dir",
        description="Delete files related to Transonic in a directory",
    )
    parser.add_argument("path", help="Path directory.")

    args = parser.parse_args()

    path_dir = Path(args.path)

    if not path_dir.is_dir():
        print("Path given is not a directory", file=sys.stderr)
        sys.exit(1)

    clean_dir(path_dir)
